#include <glib.h>
#include <string.h>
#include <gsf/gsf-infile-msole.h>
#include <gsf/gsf-infile.h>
#include <gsf/gsf-input-stdio.h>
#include <gsf/gsf-utils.h>
#include <gsf/gsf-doc-meta-data.h>
#include <gsf/gsf-msole-utils.h>
#include <stdlib.h>

struct mpp_task_fixed_meta
{
  int idx;
  int task_id;			/* Unique Task ID */
  int id;			/* Task ID used for sorting */
  int size;
  int offset;
};

struct mpp_task_var_meta
{
  int idx;
  int task_id;
  int type;
  int offset;
};

int mpp_task_fixed_meta_sort(struct mpp_task_fixed_meta *op1, struct mpp_task_fixed_meta *op2)
{
  /*g_print("<!-- fixed sort %d < %d -->\n", op1->id, op2->id);*/
  return op1->id > op2->id;
}

int mpp_task_var_meta_sort(const struct mpp_task_var_meta *op1, const struct mpp_task_var_meta *op2)
{
  /*g_print("<!-- var sort %d < %d -->\n", op1->task_id, op2->task_id);*/
  return op1->task_id > op2->task_id;
}

int mpp_task_var_meta_search(const struct mpp_task_fixed_meta *op1, const struct mpp_task_var_meta *op2)
{
  /*g_print("<!-- var search %d < %d -->\n", op1->task_id, op2->task_id);*/
  return op1->task_id > op2->task_id;
}

/*
 * Utility functions
 */
int mpp_read_byte(GsfInput *inp)
{
  const guint8* data = gsf_input_read(inp, 1, NULL);
  return GSF_LE_GET_GINT8(data);
}

int mpp_read_int(GsfInput *inp)
{
  const guint8* data = gsf_input_read(inp, 4, NULL);
  return GSF_LE_GET_GINT32(data);
}

const char *mpp_string(const void *data, int size)
{
  static char *buf = NULL;
  char *pbuf;
  int i;
  gint16 *p = (gint16*)data;
  if (buf)
    g_free(buf);
  buf = g_new0(char, (size/2+1)*6);
  pbuf = buf;
  for (i = 0;i < size/2;i ++, p++) {
    int len = g_unichar_to_utf8(GSF_LE_GET_GINT16(p), pbuf);
    pbuf += len;
  }
  *pbuf = '\0';
  return buf;
}

const char* mpp_time(const guint8* data)
{
  static char buf[32];
  int time = GSF_LE_GET_GINT16(data) / 10;
  sprintf(buf, "%02d%02d", time / 60, time % 60, 0);
  return buf;
}

const char* mpp_timestamp(const guint8* data)
{
  const int MS_EPOCH = 441651600; /* 31/12/1983 */
  static char buf[32];
  time_t days = GSF_LE_GET_GINT16(data+2);
  if (days == -1)
    return "NaN";
  days *= 24*60*60;
  int time = GSF_LE_GET_GINT16(data);
  if (time == -1)
    time = 0;
  days += time*60;
  days += MS_EPOCH;
  struct tm *timestamp = gmtime(&days);
  strftime(buf, 32, "%Y%m%dT%H%M%SZ", timestamp);
  return buf;
}

const char* mpp_blob(GsfInput *inp, int size)
{
  return (const char*)gsf_base64_encode_simple(gsf_input_read(inp, size, NULL), size);
}

const char* mpp_duration_unit(int type)
{
  const int DURATION_UNITS_MASK = 0x1F;
  switch (type) {
  case 3: return "minutes";
  case 4: return "elapsed-minutes";
  case 5: return "hours";
  case 6: return "elapsed-hours";
  case 8: return "elapsed-days";
  case 9: return "weeks";
  case 10: return "elapsed-weeks";
  case 11: return "months";
  case 12: return "elapsed-months";
  case 7: return "days";
  default: return "unknown"; /* presumably days */
  }
}

const char *mpp_constraint_type(int type)
{
  const char *types[4] = { "FF", "FS", "SF", "SS" };
  if (type >= 0 && type < 4)
    return types[type];
  return "unknown";
}

const char* mpp_constraint_type2(int type)
{
  const char *types[8] = {
    "ASAP",
    "ALAP",
    "MSO",
    "MFO",
    "SNET",
    "SNLT",
    "FNET",
    "FNLT"
  };
  if (type >= 0 && type < 8)
    return types[type];
  return "unknown";
}

static const char *xml_quote(const char *s)
{
  static char *buf = NULL;
  if (buf)
    g_free(buf);
  int i,n,c;
  const char *p;
  char *p2;
  c = 0;
  for (p = s;*p;p++)
    switch (*p) {
    case '>': c += 4; break;
    case '<': c += 4; break;
    case '&': c += 5; break;
    case '"': c += 6; break;
    case '\'': c += 6; break;
    default: c++;
    }
  buf = g_new0(char, c+1);
  for (p = s, p2 = buf;*p;p++)
    switch (*p) {
    case '>': strcpy(p2, "&gt;"); p2 += 4; break;
    case '<': strcpy(p2, "&lt;"); p2 += 4; break;
    case '&': strcpy(p2, "&amp;"); p2 += 5; break;
    case '"': strcpy(p2, "&quot;"); p2 += 6; break;
    case '\'': strcpy(p2, "&apos;"); p2 += 6; break;
    default:
      *p2++ = *p;
    }

  return buf;
}

GsfInput *mpp_follow(GsfInfile *inf, const char *path)
{
  char *sep;
  char *buf, *pbuf;
  GsfInput *inp;

  buf = pbuf = g_strdup(path);
  while ((sep = strchr(pbuf, '/')) != NULL) {
    *sep = '\0';
    inf = GSF_INFILE(gsf_infile_child_by_name(inf, pbuf));
    pbuf = sep+1;
  }
  inp = gsf_infile_child_by_name(inf, pbuf);
  g_free(buf);
  return inp;
}

/*
 * Transformation.
 * Output an XML tag with proper data transformation
 */
typedef struct _prop_transform prop_transform;
struct _prop_transform
{
  int key;
  const char *name;
  int (*transform)(const prop_transform *p, const void *data, int size, void *udata);
};

int prop_blob(const prop_transform *p, const void *data, int size, void *udata)
{
  g_print("<%s>%s</%s>\n", p->name, gsf_base64_encode_simple(data, size), p->name);
}

int var_prop_blob(const prop_transform *p, const void *data, int size, void *udata)
{
  g_print("<%s unique-id=\"%d\">%s</%s>\n", p->name, *(int*)udata, gsf_base64_encode_simple(data, size), p->name);
}

/*
 * Dummy function, for placeholder only
 */
int prop_null(const prop_transform *p, const void *data, int size, void *udata)
{
  g_print("<%s value=\"NULL\"/>\n", p->name);
}

int var_prop_null(const prop_transform *p, const void *data, int size, void *udata)
{
  g_print("<%s unique-id=\"%d\" value=\"NULL\"/>\n", p->name, *(int*)udata);
}

int prop_int(const prop_transform *p, const void *data, int size, void *udata)
{
  g_print("<%s value=\"%d\"/>\n", p->name, GSF_LE_GET_GINT32(data));
}

int var_prop_int(const prop_transform *p, const void *data, int size, void *udata)
{
  g_print("<%s unique-id=\"%d\" value=\"%d\"/>\n", p->name, *(int*)udata, GSF_LE_GET_GINT32(data));
}

int prop_short(const prop_transform *p, const void *data, int size, void *udata)
{
  g_print("<%s value=\"%d\"/>\n", p->name, GSF_LE_GET_GINT16(data));
}

int var_prop_short(const prop_transform *p, const void *data, int size, void *udata)
{
  g_print("<%s unique-id=\"%d\" value=\"%d\"/>\n", p->name, *(int*)udata, GSF_LE_GET_GINT16(data));
}

int prop_byte(const prop_transform *p, const void *data, int size, void *udata)
{
  g_print("<%s value=\"%d\"/>\n", p->name, GSF_LE_GET_GINT8(data));
}

int var_prop_byte(const prop_transform *p, const void *data, int size, void *udata)
{
  g_print("<%s unique-id=\"%d\" value=\"%d\"/>\n", p->name, *(int*)udata, GSF_LE_GET_GINT8(data));
}

int prop_bool(const prop_transform *p, const void *data, int size, void *udata)
{
  g_print("<%s value=\"%s\"/>\n", p->name, GSF_LE_GET_GINT16(data) ? "true" : "false");
}

int var_prop_bool(const prop_transform *p, const void *data, int size, void *udata)
{
  g_print("<%s unique-id=\"%d\" value=\"%s\"/>\n", p->name, *(int*)udata, GSF_LE_GET_GINT16(data) ? "true" : "false");
}

int prop_double(const prop_transform *p, const void *data, int size, void *udata)
{
  g_print("<%s value=\"%g\"/>\n", p->name, GSF_LE_GET_DOUBLE(data));
}

int var_prop_double(const prop_transform *p, const void *data, int size, void *udata)
{
  g_print("<%s unique-id=\"%d\" value=\"%g\"/>\n", p->name, *(int*)udata, GSF_LE_GET_DOUBLE(data));
}

int prop_time(const prop_transform *p, const void *data, int size, void *udata)
{
  g_print("<%s value=\"%s\"/>\n", p->name, mpp_time(data));
}

int var_prop_time(const prop_transform *p, const void *data, int size, void *udata)
{
  g_print("<%s unique-id=\"%d\" value=\"%s\"/>\n", p->name, *(int*)udata, mpp_time(data));
}

int prop_timestamp(const prop_transform *p, const void *data, int size, void *udata)
{
  g_print("<%s value=\"%s\"/>\n", p->name, mpp_timestamp(data));
}

int var_prop_timestamp(const prop_transform *p, const void *data, int size, void *udata)
{
  g_print("<%s unique-id=\"%d\" value=\"%s\"/>\n", p->name, *(int*)udata, mpp_timestamp(data));
}

int prop_string(const prop_transform *p, const void *data, int size, void *udata)
{
  g_print("<%s value=\"%s\"/>\n", p->name, xml_quote(mpp_string(data, size)));
}

int var_prop_string(const prop_transform *p, const void *data, int size, void *udata)
{
  g_print("<%s unique-id=\"%d\" value=\"%s\"/>\n", p->name, *(int*)udata, xml_quote(mpp_string(data, size)));
}

int mpp_props(GsfInput *inp)
{
  const prop_transform trans[] = {
    { 1048577, "TASK_FIELD_NAME_ALIASES", prop_blob },
    { 1048578, "RESOURCE_FIELD_NAME_ALIASES", prop_blob },
    { 37748740, "SCHEDULE_FROM", prop_short },
    { 37748750, "DEFAULT_CALENDAR_NAME", prop_string },
    { 37748752, "CURRENCY_SYMBOL", prop_string },
    { 37748753, "CURRENCY_PLACEMENT", prop_short },
    { 37748754, "CURRENCY_DIGITS", prop_short },
    { 37748757, "DURATION_UNITS", prop_short },
    { 37748758, "WORK_UNITS", prop_short },
    { 37748761, "TASK_UPDATES_RESOURCE", prop_bool },
    { 37748762, "SPLIT_TASKS", prop_bool },
    { 37748764, "START_TIME", prop_time },
    { 37748765, "MINUTES_PER_DAY", prop_int },
    { 37748766, "MINUTES_PER_WEEK", prop_int },
    { 37748767, "STANDARD_RATE", prop_double },
    { 37748768, "OVERTIME_RATE", prop_double },
    { 37748769, "END_TIME", prop_time },
    { 37748793, "CALCULATE_MULTIPLE_CRITICAL_PATHS", prop_bool },
    { 37748805, "STATUS_DATE", prop_timestamp },
    { 37748868, "SUBPROJECT_COUNT", prop_short },
    { 37748898, "SUBPROJECT_DATA", prop_blob },
    { 37748900, "SUBPROJECT_TASK_COUNT", prop_short },
    { 37753744, "GRAPHICAL_INDICATOR_DATA", prop_short },
    { 54525952, "FONT_BASES", prop_blob },
    { 893386752, "PASSWORD_FLAG", prop_byte },
    { 893386767, "AUTO_FILTER", prop_bool },
    { 0, NULL, NULL }
  };

  int i,n;
  /* props9 header 16 bytes */
  const guint8 *data = gsf_input_read(inp, 16, NULL);
  n = GSF_LE_GET_GINT16(data+12);

  g_print("<props count=\"%d\">\n", n);
  i = 0;
  while (i < n) {
    int size = mpp_read_int(inp);
    int key = mpp_read_int(inp);
    int attr3 = mpp_read_int(inp);

    if (gsf_input_remaining(inp) < size || size < 1)
      break;
    const guint8* data = gsf_input_read(inp, size, NULL);
    const prop_transform *tran;
    for (tran = trans; tran->key != 0; tran ++)
      if (tran->key == key) {
	tran->transform(tran, data, size, NULL);
	break;
      }
    i ++;
    if (size % 2) gsf_input_read(inp, 1, NULL);
  }
  g_print("</props>\n");
}

#define TAG_INT(name, data) g_print("<%s value=\"%d\"/>\n", name, GSF_LE_GET_GINT32(data))
#define TAG_SHORT(name, data) g_print("<%s value=\"%d\"/>\n", name, GSF_LE_GET_GINT16(data))
#define TAG_SHORT_MAP(name, data, func) g_print("<%s value=\"%s\"/>\n", name, func(GSF_LE_GET_GINT16(data)))
#define TAG_DOUBLE(name, data) g_print("<%s value=\"%g\"/>\n", name, GSF_LE_GET_DOUBLE(data))
#define TAG_DOUBLE2(name, data, scale) g_print("<%s value=\"%g\"/>\n", name, GSF_LE_GET_DOUBLE(data)/scale)
#define TAG_STAMP(name, data) g_print("<%s value=\"%s\"/>\n", name, mpp_timestamp(data))
#define TAG_DURATION(name, data, unit) g_print("<%s value=\"%d\" unit=\"%s\"/>\n", name, GSF_LE_GET_GINT32(data), unit)
#define TAG_DURATION2(name, data, scale, unit) g_print("<%s value=\"%d\" unit=\"%s\"/>\n", name, GSF_LE_GET_GINT32(data)/scale, unit)
#define TAG_DURATION_MAP(name, data, unit) TAG_DURATION(name, data, mpp_duration_unit(unit))
#define TAG_RATE(name, data, unit) /* FIXME */

int mpp_task_fixed_data(const guint8 *data, int offset, int size)
{
  const char *duration_unit = mpp_duration_unit(GSF_LE_GET_GINT16(data+64));

  TAG_INT("UniqueId", data);
  TAG_DOUBLE2("ActualCost", data+216, 100);
  TAG_DURATION("ActualDuration", data+66, duration_unit);
  TAG_STAMP("ActualFinish", data+100);
  TAG_STAMP("ActualStart", data+96);
  TAG_DOUBLE2("ActualWork", data+184, 60000);
  TAG_DOUBLE2("BaselineCost", data+232, 100);
  TAG_SHORT("BaselineDuration", data+78);
  TAG_STAMP("BaselineFinish", data+108);
  TAG_STAMP("BaselineStart", data+104);
  TAG_DOUBLE2("BaselineWork", data+176, 60000);
  TAG_STAMP("ConstraintDate", data+112);
  TAG_SHORT_MAP("ConstraintType", data+80, mpp_constraint_type2);
  TAG_DOUBLE2("Cost", data+200, 100);
  TAG_STAMP("CreateDate", data+130);
  TAG_STAMP("Deadline", data+164);
  TAG_DURATION("Duration", data+60, duration_unit);
  TAG_STAMP("EarlyFinish", data+8);
  TAG_STAMP("EarlyStart", data+88);
  g_print("<Estimated value=\"%d\"/>\n", duration_unit); /* FIXME */
  TAG_STAMP("Finish", data+8);
  TAG_DOUBLE2("FixedCost", data+208, 100);
  TAG_SHORT("FixedCostAccrual", data+128);
  TAG_DURATION("FreeSlack", data+24, duration_unit);
  TAG_INT("ID", data+4);
  TAG_STAMP("LateFinish", data+92);
  TAG_STAMP("LateStart", data+148);
  TAG_SHORT("LevelingDelay", data+86);
  TAG_SHORT("OutlineLevel", data+40);
  TAG_SHORT("PercentageComplete", data+122);
  TAG_SHORT("PercentageWorkComplete", data+124);
  TAG_STAMP("PreleveledFinish", data+140);
  TAG_STAMP("PreleveledStart", data+136);
  TAG_SHORT("Priority", data+120);
  TAG_DOUBLE2("RemainingCost", data+224, 100);
  TAG_DURATION("RemainingDuration", data+70, duration_unit);
  TAG_DOUBLE2("RemainingWork", data+192, 60000);
  TAG_STAMP("Resume", data+20);
  TAG_STAMP("Start", data+88);
  TAG_DURATION("StartSlack", data+28, duration_unit);
  TAG_STAMP("Stop", data+16);
  TAG_SHORT("Type", data+126);
  TAG_DOUBLE2("Work", data+168, 60000);
  TAG_DURATION("FinishSlack", data+32, duration_unit);
  if (GSF_LE_GET_GINT32(data+160) != -1)
    TAG_INT("Calendar", data+160);
}

int mpp_task_var_data(struct mpp_task_var_meta *var_meta, GsfInput *inp2)
{
  static const prop_transform trans[] = {
    { 3, "TASK_ACTUAL_OVERTIME_WORK", var_prop_null },
    { 4, "TASK_REMAINING_OVERTIME_WORK", var_prop_null },
    { 5, "TASK_OVERTIME_COST", var_prop_null },
    { 6, "TASK_ACTUAL_OVERTIME_COST", var_prop_null },
    { 7, "TASK_REMAINING_OVERTIME_COST", var_prop_null },

    { 8, "TASK_SUBPROJECT_TASKS_UNIQUEID_OFFSET", var_prop_int },
    { 9, "TASK_SUBPROJECTTASKID", var_prop_int },

    { 10, "TASK_WBS", var_prop_string },
    { 11, "TASK_NAME", var_prop_string },
    { 12, "TASK_CONTACT", var_prop_string },

    { 14, "TASK_TEXT1", var_prop_string },
    { 15, "TASK_TEXT2", var_prop_string },
    { 16, "TASK_TEXT3", var_prop_string },
    { 17, "TASK_TEXT4", var_prop_string },
    { 18, "TASK_TEXT5", var_prop_string },
    { 19, "TASK_TEXT6", var_prop_string },
    { 20, "TASK_TEXT7", var_prop_string },
    { 21, "TASK_TEXT8", var_prop_string },
    { 22, "TASK_TEXT9", var_prop_string },
    { 23, "TASK_TEXT10", var_prop_string },

    { 24, "TASK_START1", var_prop_timestamp },
    { 25, "TASK_FINISH1", var_prop_timestamp },
    { 26, "TASK_START2", var_prop_timestamp },
    { 27, "TASK_FINISH2", var_prop_timestamp },
    { 28, "TASK_START3", var_prop_timestamp },
    { 29, "TASK_FINISH3", var_prop_timestamp },
    { 30, "TASK_START4", var_prop_timestamp },
    { 31, "TASK_FINISH4", var_prop_timestamp },
    { 32, "TASK_START5", var_prop_timestamp },
    { 33, "TASK_FINISH5", var_prop_timestamp },
    { 34, "TASK_START6", var_prop_timestamp },
    { 35, "TASK_FINISH6", var_prop_timestamp },
    { 36, "TASK_START7", var_prop_timestamp },
    { 37, "TASK_FINISH7", var_prop_timestamp },
    { 38, "TASK_START8", var_prop_timestamp },
    { 39, "TASK_FINISH8", var_prop_timestamp },
    { 40, "TASK_START9", var_prop_timestamp },
    { 41, "TASK_FINISH9", var_prop_timestamp },
    { 42, "TASK_START10", var_prop_timestamp },
    { 43, "TASK_FINISH10", var_prop_timestamp },

    { 45, "TASK_NUMBER1", var_prop_double },
    { 46, "TASK_NUMBER2", var_prop_double },
    { 47, "TASK_NUMBER3", var_prop_double },
    { 48, "TASK_NUMBER4", var_prop_double },
    { 49, "TASK_NUMBER5", var_prop_double },
    { 50, "TASK_NUMBER6", var_prop_double },
    { 51, "TASK_NUMBER7", var_prop_double },
    { 52, "TASK_NUMBER8", var_prop_double },
    { 53, "TASK_NUMBER9", var_prop_double },
    { 54, "TASK_NUMBER10", var_prop_double },

    { 55, "TASK_DURATION1", var_prop_null },
    { 56, "TASK_DURATION1_UNITS", var_prop_null },
    { 57, "TASK_DURATION2", var_prop_null },
    { 58, "TASK_DURATION2_UNITS", var_prop_null },
    { 59, "TASK_DURATION3", var_prop_null },
    { 60, "TASK_DURATION3_UNITS", var_prop_null },
    { 61, "TASK_DURATION4", var_prop_null },
    { 62, "TASK_DURATION4_UNITS", var_prop_null },
    { 63, "TASK_DURATION5", var_prop_null },
    { 64, "TASK_DURATION5_UNITS", var_prop_null },
    { 65, "TASK_DURATION6", var_prop_null },
    { 66, "TASK_DURATION6_UNITS", var_prop_null },
    { 67, "TASK_DURATION7", var_prop_null },
    { 68, "TASK_DURATION7_UNITS", var_prop_null },
    { 69, "TASK_DURATION8", var_prop_null },
    { 70, "TASK_DURATION8_UNITS", var_prop_null },
    { 71, "TASK_DURATION9", var_prop_null },
    { 72, "TASK_DURATION9_UNITS", var_prop_null },
    { 73, "TASK_DURATION10", var_prop_null },
    { 74, "TASK_DURATION10_UNITS", var_prop_null },

    { 79, "TASK_EXTERNAL_TASK_ID", var_prop_int },

    { 80, "TASK_DATE1", var_prop_timestamp },
    { 81, "TASK_DATE2", var_prop_timestamp },
    { 82, "TASK_DATE3", var_prop_timestamp },
    { 83, "TASK_DATE4", var_prop_timestamp },
    { 84, "TASK_DATE5", var_prop_timestamp },
    { 85, "TASK_DATE6", var_prop_timestamp },
    { 86, "TASK_DATE7", var_prop_timestamp },
    { 87, "TASK_DATE8", var_prop_timestamp },
    { 88, "TASK_DATE9", var_prop_timestamp },
    { 89, "TASK_DATE10", var_prop_timestamp },

    { 90, "TASK_TEXT11", var_prop_string },
    { 91, "TASK_TEXT12", var_prop_string },
    { 92, "TASK_TEXT13", var_prop_string },
    { 93, "TASK_TEXT14", var_prop_string },
    { 94, "TASK_TEXT15", var_prop_string },
    { 95, "TASK_TEXT16", var_prop_string },
    { 96, "TASK_TEXT17", var_prop_string },
    { 97, "TASK_TEXT18", var_prop_string },
    { 98, "TASK_TEXT19", var_prop_string },
    { 99, "TASK_TEXT20", var_prop_string },
    { 100, "TASK_TEXT21", var_prop_string },
    { 101, "TASK_TEXT22", var_prop_string },
    { 102, "TASK_TEXT23", var_prop_string },
    { 103, "TASK_TEXT24", var_prop_string },
    { 104, "TASK_TEXT25", var_prop_string },
    { 105, "TASK_TEXT26", var_prop_string },
    { 106, "TASK_TEXT27", var_prop_string },
    { 107, "TASK_TEXT28", var_prop_string },
    { 108, "TASK_TEXT29", var_prop_string },
    { 109, "TASK_TEXT30", var_prop_string },

    { 110, "TASK_NUMBER11", var_prop_double },
    { 111, "TASK_NUMBER12", var_prop_double },
    { 112, "TASK_NUMBER13", var_prop_double },
    { 113, "TASK_NUMBER14", var_prop_double },
    { 114, "TASK_NUMBER15", var_prop_double },
    { 115, "TASK_NUMBER16", var_prop_double },
    { 116, "TASK_NUMBER17", var_prop_double },
    { 117, "TASK_NUMBER18", var_prop_double },
    { 118, "TASK_NUMBER19", var_prop_double },
    { 119, "TASK_NUMBER20", var_prop_double },

    { 123, "TASK_OUTLINECODE1", var_prop_int },
    { 124, "TASK_OUTLINECODE2", var_prop_int },
    { 125, "TASK_OUTLINECODE3", var_prop_int },
    { 126, "TASK_OUTLINECODE4", var_prop_int },
    { 127, "TASK_OUTLINECODE5", var_prop_int },
    { 128, "TASK_OUTLINECODE6", var_prop_int },
    { 129, "TASK_OUTLINECODE7", var_prop_int },
    { 130, "TASK_OUTLINECODE8", var_prop_int },
    { 131, "TASK_OUTLINECODE9", var_prop_int },
    { 132, "TASK_OUTLINECODE10", var_prop_int },

    { 133, "TASK_HYPERLINK", var_prop_null },

    { 134, "TASK_COST1", var_prop_double },
    { 135, "TASK_COST2", var_prop_double },
    { 136, "TASK_COST3", var_prop_double },
    { 137, "TASK_COST4", var_prop_double },
    { 138, "TASK_COST5", var_prop_double },
    { 139, "TASK_COST6", var_prop_double },
    { 140, "TASK_COST7", var_prop_double },
    { 141, "TASK_COST8", var_prop_double },
    { 142, "TASK_COST9", var_prop_double },
    { 143, "TASK_COST10", var_prop_double },

    { 144, "TASK_NOTES", var_prop_string },
    { 0, NULL, NULL }
  };

  int unique_id = var_meta->task_id;
  int type = var_meta->type;
  int offset = var_meta->offset;

  gsf_input_seek(inp2, offset, G_SEEK_SET);
  int size = mpp_read_int(inp2);
  const guint8* data = gsf_input_read(inp2, size, NULL);
  /*g_print("<item task-id=\"%d\" type=\"%d\" offset=\"%d\" size=\"%d\">\n", unique_id, type, offset, size);*/

  const prop_transform *tran;
  for (tran = trans; tran->key != 0; tran ++)
    if (tran->key == type) {
      tran->transform(tran, data, size, &unique_id);
      break;
    }

  /*g_print("</item>\n");*/
}


int mpp_resource_fixed_data(const guint8 *data)
{
  TAG_SHORT("AccrueAt", data+12);
  TAG_DOUBLE2("ActualCost", data+132, 100);
  TAG_DOUBLE2("ActualOvertimeCost", data+132, 100);
  TAG_DURATION2("ActualWork", data+60, 60000, "hours");
  TAG_STAMP("AvailableFrom", data+20);
  TAG_STAMP("AvailableTo", data+24);
  TAG_DOUBLE2("BaselineCost", data+148, 100);
  TAG_DURATION2("BaselineWork", data+68, 60000, "hours");
  TAG_DOUBLE2("Cost", data+140, 100);
  TAG_DOUBLE2("CostPerUse", data+84, 100);
  TAG_DOUBLE2("MaxUnits", data+44, 100);
  TAG_RATE("OvertimeRate", data+36, "hours");
  TAG_DURATION2("OvertimeWork", data+76, 60000, "hours");
  TAG_DOUBLE2("PeakUnits", data+124, 100);
  TAG_DURATION2("RegularWork", data+100, 60000, "hours");
  TAG_DOUBLE2("RemainingCost", data+156, 100);
  TAG_DOUBLE2("RemainingOvertimeCost", data+180, 100);
  TAG_DURATION2("RemainingWork", data+92, 60000, "hours");
  TAG_RATE("StandardRate", data+28, "hours");
  g_printf("<Type value=\"%s\"/>\n", GSF_LE_GET_GINT16(data+14) == 0 ? "work" : "material");
  TAG_INT("UniqueId", data);
  TAG_INT("ID", data+4);
  TAG_DURATION2("Work", data+52, 60000, "hours");
}

int mpp_resource_var_data(GsfInput *inp2, int type, int unique_id)
{
  static const prop_transform trans[] = {
    { 1, "RESOURCE_NAME", var_prop_string },
    { 3, "RESOURCE_INITIALS", var_prop_string },
    { 4, "RESOURCE_GROUP", var_prop_string },
    { 5, "RESOURCE_CODE", var_prop_string },
    { 6, "RESOURCE_EMAIL", var_prop_string },

    { 10, "RESOURCE_TEXT1", var_prop_string },
    { 11, "RESOURCE_TEXT2", var_prop_string },
    { 12, "RESOURCE_TEXT3", var_prop_string },
    { 13, "RESOURCE_TEXT4", var_prop_string },
    { 14, "RESOURCE_TEXT5", var_prop_string },
    { 15, "RESOURCE_TEXT6", var_prop_string },
    { 16, "RESOURCE_TEXT7", var_prop_string },
    { 17, "RESOURCE_TEXT8", var_prop_string },
    { 18, "RESOURCE_TEXT9", var_prop_string },
    { 19, "RESOURCE_TEXT10", var_prop_string },
    { 20, "RESOURCE_TEXT11", var_prop_string },
    { 21, "RESOURCE_TEXT12", var_prop_string },
    { 22, "RESOURCE_TEXT13", var_prop_string },
    { 23, "RESOURCE_TEXT14", var_prop_string },
    { 24, "RESOURCE_TEXT15", var_prop_string },
    { 25, "RESOURCE_TEXT16", var_prop_string },
    { 26, "RESOURCE_TEXT17", var_prop_string },
    { 27, "RESOURCE_TEXT18", var_prop_string },
    { 28, "RESOURCE_TEXT19", var_prop_string },
    { 29, "RESOURCE_TEXT20", var_prop_string },
    { 30, "RESOURCE_TEXT21", var_prop_string },
    { 31, "RESOURCE_TEXT22", var_prop_string },
    { 32, "RESOURCE_TEXT23", var_prop_string },
    { 33, "RESOURCE_TEXT24", var_prop_string },
    { 34, "RESOURCE_TEXT25", var_prop_string },
    { 35, "RESOURCE_TEXT26", var_prop_string },
    { 36, "RESOURCE_TEXT27", var_prop_string },
    { 37, "RESOURCE_TEXT28", var_prop_string },
    { 38, "RESOURCE_TEXT29", var_prop_string },
    { 39, "RESOURCE_TEXT30", var_prop_string },

    { 40, "RESOURCE_START1", var_prop_timestamp },
    { 41, "RESOURCE_START2", var_prop_timestamp },
    { 42, "RESOURCE_START3", var_prop_timestamp },
    { 43, "RESOURCE_START4", var_prop_timestamp },
    { 44, "RESOURCE_START5", var_prop_timestamp },
    { 45, "RESOURCE_START6", var_prop_timestamp },
    { 46, "RESOURCE_START7", var_prop_timestamp },
    { 47, "RESOURCE_START8", var_prop_timestamp },
    { 48, "RESOURCE_START9", var_prop_timestamp },
    { 49, "RESOURCE_START10", var_prop_timestamp },

    { 50, "RESOURCE_FINISH1", var_prop_timestamp },
    { 51, "RESOURCE_FINISH2", var_prop_timestamp },
    { 52, "RESOURCE_FINISH3", var_prop_timestamp },
    { 53, "RESOURCE_FINISH4", var_prop_timestamp },
    { 54, "RESOURCE_FINISH5", var_prop_timestamp },
    { 55, "RESOURCE_FINISH6", var_prop_timestamp },
    { 56, "RESOURCE_FINISH7", var_prop_timestamp },
    { 57, "RESOURCE_FINISH8", var_prop_timestamp },
    { 58, "RESOURCE_FINISH9", var_prop_timestamp },
    { 59, "RESOURCE_FINISH10", var_prop_timestamp },

    { 60, "RESOURCE_NUMBER1", var_prop_double },
    { 61, "RESOURCE_NUMBER2", var_prop_double },
    { 62, "RESOURCE_NUMBER3", var_prop_double },
    { 63, "RESOURCE_NUMBER4", var_prop_double },
    { 64, "RESOURCE_NUMBER5", var_prop_double },
    { 65, "RESOURCE_NUMBER6", var_prop_double },
    { 66, "RESOURCE_NUMBER7", var_prop_double },
    { 67, "RESOURCE_NUMBER8", var_prop_double },
    { 68, "RESOURCE_NUMBER9", var_prop_double },
    { 69, "RESOURCE_NUMBER10", var_prop_double },
    { 70, "RESOURCE_NUMBER11", var_prop_double },
    { 71, "RESOURCE_NUMBER12", var_prop_double },
    { 72, "RESOURCE_NUMBER13", var_prop_double },
    { 73, "RESOURCE_NUMBER14", var_prop_double },
    { 74, "RESOURCE_NUMBER15", var_prop_double },
    { 75, "RESOURCE_NUMBER16", var_prop_double },
    { 76, "RESOURCE_NUMBER17", var_prop_double },
    { 77, "RESOURCE_NUMBER18", var_prop_double },
    { 78, "RESOURCE_NUMBER19", var_prop_double },
    { 79, "RESOURCE_NUMBER20", var_prop_double },

    { 80, "RESOURCE_DURATION1", var_prop_int },
    { 81, "RESOURCE_DURATION2", var_prop_int },
    { 82, "RESOURCE_DURATION3", var_prop_int },
    { 83, "RESOURCE_DURATION4", var_prop_int },
    { 84, "RESOURCE_DURATION5", var_prop_int },
    { 85, "RESOURCE_DURATION6", var_prop_int },
    { 86, "RESOURCE_DURATION7", var_prop_int },
    { 87, "RESOURCE_DURATION8", var_prop_int },
    { 88, "RESOURCE_DURATION9", var_prop_int },
    { 89, "RESOURCE_DURATION10", var_prop_int },

    { 90, "RESOURCE_DURATION1_UNITS", var_prop_short },
    { 91, "RESOURCE_DURATION2_UNITS", var_prop_short },
    { 92, "RESOURCE_DURATION3_UNITS", var_prop_short },
    { 93, "RESOURCE_DURATION4_UNITS", var_prop_short },
    { 94, "RESOURCE_DURATION5_UNITS", var_prop_short },
    { 95, "RESOURCE_DURATION6_UNITS", var_prop_short },
    { 96, "RESOURCE_DURATION7_UNITS", var_prop_short },
    { 97, "RESOURCE_DURATION8_UNITS", var_prop_short },
    { 98, "RESOURCE_DURATION9_UNITS", var_prop_short },
    { 99, "RESOURCE_DURATION10_UNITS", var_prop_short },

    { 102, "RESOURCE_SUBPROJECTRESOURCEID", var_prop_null },

    { 103, "RESOURCE_DATE1", var_prop_timestamp },
    { 104, "RESOURCE_DATE2", var_prop_timestamp },
    { 105, "RESOURCE_DATE3", var_prop_timestamp },
    { 106, "RESOURCE_DATE4", var_prop_timestamp },
    { 107, "RESOURCE_DATE5", var_prop_timestamp },
    { 108, "RESOURCE_DATE6", var_prop_timestamp },
    { 109, "RESOURCE_DATE7", var_prop_timestamp },
    { 110, "RESOURCE_DATE8", var_prop_timestamp },
    { 111, "RESOURCE_DATE9", var_prop_timestamp },
    { 112, "RESOURCE_DATE10", var_prop_timestamp },

    { 113, "RESOURCE_OUTLINECODE1", var_prop_null },
    { 114, "RESOURCE_OUTLINECODE2", var_prop_null },
    { 115, "RESOURCE_OUTLINECODE3", var_prop_null },
    { 116, "RESOURCE_OUTLINECODE4", var_prop_null },
    { 117, "RESOURCE_OUTLINECODE5", var_prop_null },
    { 118, "RESOURCE_OUTLINECODE6", var_prop_null },
    { 119, "RESOURCE_OUTLINECODE7", var_prop_null },
    { 120, "RESOURCE_OUTLINECODE8", var_prop_null },
    { 121, "RESOURCE_OUTLINECODE9", var_prop_null },
    { 122, "RESOURCE_OUTLINECODE10", var_prop_null },

    { 124, "RESOURCE_NOTES", var_prop_string },

    { 125, "RESOURCE_COST1", var_prop_double },
    { 126, "RESOURCE_COST2", var_prop_double },
    { 127, "RESOURCE_COST3", var_prop_double },
    { 128, "RESOURCE_COST4", var_prop_double },
    { 129, "RESOURCE_COST5", var_prop_double },
    { 130, "RESOURCE_COST6", var_prop_double },
    { 131, "RESOURCE_COST7", var_prop_double },
    { 132, "RESOURCE_COST8", var_prop_double },
    { 133, "RESOURCE_COST9", var_prop_double },
    { 134, "RESOURCE_COST10", var_prop_double },
    { 0, NULL, NULL }
  };

  int size = mpp_read_int(inp2);
  const guint8* data = gsf_input_read(inp2, size, NULL);

  const prop_transform *tran;
  for (tran = trans; tran->key != 0; tran ++)
    if (tran->key == type) {
      tran->transform(tran, data, size, &unique_id);
      break;
    }

}

int mpp_assignment_fixed_data(const guint8 *data)
{
  const char *duration_unit = mpp_duration_unit(GSF_LE_GET_GINT16(data+64));

  TAG_INT("Id", data);
  TAG_INT("TaskId", data+4);
  TAG_INT("ResourceId", data+8);
  TAG_DOUBLE2("ActualCost", data+110, 100);
  TAG_DURATION2("ActualWork", data+70, 100, "hours");
  TAG_DOUBLE2("Cost", data+102, 100);
  TAG_DURATION("Delay", data+24, "hours");
  TAG_STAMP("Finish", data+16);
  TAG_DURATION2("RemainingWork", data+86, 100, "hours");
  TAG_STAMP("Start", data+12);
  TAG_DOUBLE2("Units", data+54, 100);
  TAG_DURATION2("Work", data+62, 100, "hours");
}

int mpp_assignment_complete_hours(const guint8 *data, int unique_id)
{
  int i, split_count = GSF_LE_GET_GINT16(data);
  g_print("<complete-hours unique-id=\"%d\" count=\"%d\">\n", unique_id, split_count);
  if (split_count) {
    TAG_DURATION2("split-duration", data+24, 4800, "hours");
    for (i = 0;i < split_count;i ++) {
      int split_time = GSF_LE_GET_GINT32(data+32+i*20);
      g_print("<split value=\"%g\" unit=\"hours\"/>\n", (double)split_time/4800);
    }
  }
  g_print("</complete-hours>\n", split_count);
}

int mpp_assignment_incomplete_hours(const guint8 *data, int unique_id)
{
  int i, split_count = GSF_LE_GET_GINT16(data);
  g_print("<incomplete-hours unique-id=\"%d\" count=\"%d\">\n", unique_id, split_count);
  if (split_count) {
    int time_offset = 0; /* FIXME */
    for (i = 0;i < split_count;i ++) {
      int split_time = GSF_LE_GET_GINT32(data+24+44+i*28);
      g_print("<split value=\"%g\" unit=\"hours\"/>\n", (double)split_time/4800);
    }
  }
  else {
    int time_offset = 0; /* FIXME */
    int split_time = GSF_LE_GET_GINT32(data+24);
    g_print("<split value=\"%g\" unit=\"hours\"/>\n", (double)split_time/4800);
  }
  g_print("</incomplete-hours>\n", split_count);
}


struct mpp_task_fixed_meta* mpp_task_fixed_meta_readahead(GsfInput *inp, GsfInput *inp2, int itemSize, int dataSize)
{
  int avail = gsf_input_remaining(inp);
  int magic = mpp_read_int(inp);
  if (magic != 0xFADFADBA) {
    g_print("<!-- BAD MAGIC -->\n");
    return NULL;
  }
  mpp_read_int(inp); /* unknown */
  int item_count = mpp_read_int(inp);
  mpp_read_int(inp); /* unknown */
  item_count = (avail - 16) / itemSize;
  int i;
  struct mpp_task_fixed_meta *fixed_meta = g_new0(struct mpp_task_fixed_meta, item_count+1);
  fixed_meta[item_count].idx = -1;
  /* int inp2_size = gsf_input_remaining(inp2); */
  for (i = 0;i < item_count;i ++) {
    const guint8 *data = gsf_input_read(inp, itemSize, NULL);
    int size = fixed_meta[i].size = GSF_LE_GET_GINT32(data);
    int offset = fixed_meta[i].offset = GSF_LE_GET_GINT32(data+4);
    fixed_meta[i].idx = i;

    if (offset > gsf_input_size(inp2))
      continue;

    if (size < 240) /* MINIMUM_EXPECTED_TASK_SIZE */
      continue;

    gsf_input_seek(inp2, offset, G_SEEK_SET);
    fixed_meta[i].task_id = mpp_read_int(inp2);
    fixed_meta[i].id = mpp_read_int(inp2); /* +4 */
  }
  return fixed_meta;
}

struct mpp_task_var_meta* mpp_task_var_meta_readahead(GsfInput *inp)
{
  int magic, i, item_count, data_size;
  struct mpp_task_var_meta *var_meta;
  guint8 buf[4];

  magic = mpp_read_int(inp);
  if (magic != 0xFADFADBA) {
    g_print("<!-- BAD MAGIC -->\n");
    return NULL;
  }
  mpp_read_int(inp); /* unknown */
  item_count = mpp_read_int(inp);
  mpp_read_int(inp); /* unknown */
  mpp_read_int(inp); /* unknown */
  data_size = mpp_read_int(inp);

  var_meta = g_new0(struct mpp_task_var_meta, item_count+1);
  var_meta[item_count].idx = -1;

  for (i = 0;i < item_count;i ++) {
    var_meta[i].idx = i;
    gsf_input_read(inp, 3, buf);
    buf[3] = 0;
    var_meta[i].task_id = GSF_LE_GET_GINT32(buf);
    var_meta[i].type = (unsigned char)mpp_read_byte(inp);
    var_meta[i].offset = mpp_read_int(inp);
  }
  return var_meta;
}

int mpp_task(struct mpp_task_fixed_meta *fixed_meta,
	      GsfInput *inp_fixed_meta, int itemSize,
	      GsfInput *inp_fixed_data, int maxExpectedDataSize,
	      struct mpp_task_var_meta *var_meta,
	      GsfInput *inp_var_data)
{
  int i, n_var_meta;
  int n_closing = 0;
  int previous_outline_level = 0, outline_level = 0;

  for (n_var_meta = 0;var_meta[n_var_meta].idx != -1;n_var_meta ++);

  g_print("<tasks>\n");

  for (i = 0;fixed_meta[i].idx != -1;i ++) {
    gsf_input_seek(inp_fixed_meta, 16 + fixed_meta[i].idx*itemSize, G_SEEK_SET);
    const guint8 *data = gsf_input_read(inp_fixed_meta, itemSize, NULL);
    int offset = fixed_meta[i].offset;
    int size = fixed_meta[i].size;
    int is_effort_driven = data[11] & 0x10;
    int is_expanded = data[12] & 0x02;
    int set_hide_bar = data[10] & 0x80;
    int set_level_assignments = data[13] & 0x04;
    int set_level_can_split = data[13] & 0x02;
    int is_marked = data[9] & 0x40;
    int is_milestone = data[8] & 0x20;
    int set_rollup = data[10] & 0x08;
    int set_split = data[9] & 0x80;

    if (offset > gsf_input_size(inp_fixed_data))
      continue;

    if (size < 240) /* MINIMUM_EXPECTED_TASK_SIZE */
      continue;

    int avail = gsf_input_size(inp_fixed_data) - offset;
    if (size < 0 || size > avail)
      size = maxExpectedDataSize > 0 && maxExpectedDataSize < avail ? maxExpectedDataSize : avail;
    gsf_input_seek(inp_fixed_data, offset, G_SEEK_SET);
    data = gsf_input_read(inp_fixed_data, size, NULL);
    outline_level = GSF_LE_GET_GINT16(data+40); /* see mpp_task_fixed_data */

    if (outline_level <= previous_outline_level)
      while (n_closing > outline_level) {
	g_print("</task>\n");
	n_closing --;
      }
    previous_outline_level = outline_level;

    g_print("<task idx=\"%d\">\n", fixed_meta[i].idx);
    n_closing ++;

    mpp_task_fixed_data(data, offset, size);

#define PROP(name,var) g_print("<" name " value=\"%s\"/>\n", (var) ? "true" : "false")
    PROP("effort-driven", is_effort_driven);
    PROP("expanded", is_expanded);
    PROP("hide-bar", set_hide_bar);
    PROP("level-assignments", set_level_assignments);
    PROP("level-can-split", set_level_can_split);
    PROP("marked", is_marked);
    PROP("milestone", is_milestone);
    PROP("rollup", set_rollup);
    PROP("split", set_split);
#undef PROP

    /*
    struct mpp_task_var_meta *var_meta_item = bsearch(fixed_meta+i, var_meta, n_var_meta, sizeof(struct mpp_task_var_meta), var_meta_search);
    */
    struct mpp_task_var_meta *var_meta_item;
    for (var_meta_item = var_meta; var_meta_item->idx != -1; var_meta_item ++)
      if (var_meta_item->task_id == fixed_meta[i].task_id) {
	while (var_meta_item->task_id == fixed_meta[i].task_id) {
	  mpp_task_var_data(var_meta_item, inp_var_data);
	  var_meta_item++;
	}
	break;
      }
  }
  while (n_closing) {
      g_print("</task>\n");
      n_closing--;
  }
  g_print("</tasks>\n");
}

int mpp_constraint(GsfInput *inp,GsfInput *inp2, int itemSize, int dataSize)
{
  int avail = gsf_input_remaining(inp);
  int magic = mpp_read_int(inp);
  if (magic != 0xFADFADBA) {
    g_print("<!-- BAD MAGIC -->\n");
    return -1;
  }
  mpp_read_int(inp); /* unknown */
  int item_count = mpp_read_int(inp);
  mpp_read_int(inp); /* unknown */
  item_count = (avail - 16) / itemSize;
  g_print("<constraints count=\"%d\">\n", item_count);
  int i;
  /* int inp2_size = gsf_input_remaining(inp2); */
  for (i = 0;i < item_count;i ++) {
    const guint8 *meta_data = gsf_input_read(inp, itemSize, NULL);
    int first_val = GSF_LE_GET_GINT32(meta_data);
    if (first_val)
      continue;

    int offset = GSF_LE_GET_GINT32(meta_data+4);
    if (offset > gsf_input_size(inp2))
      continue;

    int avail = gsf_input_size(inp2) - offset;
    int size = dataSize;
    gsf_input_seek(inp2, offset, G_SEEK_SET);
    const guint8 *data = gsf_input_read(inp2, size, NULL);
    int constraint_id = GSF_LE_GET_GINT32(data);
    int task1 = GSF_LE_GET_GINT32(data+4);
    int task2 = GSF_LE_GET_GINT32(data+8);
    int type = GSF_LE_GET_GINT16(data+12);
    int duration_unit = GSF_LE_GET_GINT16(data+14);
    int duration = GSF_LE_GET_GINT32(data+16);
    g_print("<constraint offset=\"%d\" constraint-id=\"%d\" type=\"%s\" task1=\"%d\" task2=\"%d\" duration=\"%d\" duration-unit=\"%d\"/>\n", offset, constraint_id, mpp_constraint_type(type), task1, task2, duration, duration_unit);
  }
  g_print("</constraints>\n");
}

void mpp_summary_print(char const *name, GsfDocProp const *prop)
{
  const char *link_name = gsf_doc_prop_get_link (prop);
  GValue const *val = gsf_doc_prop_get_val  (prop);
  char *tmp = g_strdup_value_contents(val); /* FIXME */
  g_print("<item name=\"%s\" ", name);
  if (link_name)
    g_print("link=\"%s\" ", link_name);
  g_print("value=\"%s\"/>\n", xml_quote(tmp));
  g_free(tmp);
}

int mpp_summary(GsfInput *inp)
{
  GsfDocMetaData *data = gsf_doc_meta_data_new();
  gsf_msole_metadata_read(inp, data);
  gsf_doc_meta_data_foreach (data, (GHFunc) mpp_summary_print, NULL);
  g_print("<!-- \n");gsf_doc_meta_dump(data);g_print("-->\n"); /* FIXME */
  g_object_unref(data);
}

int mpp_resource_fixed(GsfInput *inp, int itemSize, GsfInput *inp2)
{
  int avail = gsf_input_remaining(inp);
  int magic = mpp_read_int(inp);
  int i, item_count;
  if (magic != 0xFADFADBA) {
    g_print("<!-- BAD MAGIC -->\n");
    return -1;
  }
  mpp_read_int(inp); /* unknown */
  item_count = mpp_read_int(inp);
  mpp_read_int(inp); /* unknown */
  item_count = (avail - 16) / itemSize;
  for (i = 0;i < item_count;i ++) {
    const guint8 *data = gsf_input_read(inp, itemSize, NULL);
    int size = GSF_LE_GET_GINT32(data);
    int offset = GSF_LE_GET_GINT32(data+4);

    if (size < 188) /* MINIMUM_EXPECTED_RESOURCE_SIZE */
      continue;

    gsf_input_seek(inp2, offset, G_SEEK_SET);
    data = gsf_input_read(inp2, size, NULL);
    if (data) {
      g_print("<resource idx=\"%d\">\n", i);
      mpp_resource_fixed_data(data);
      g_print("</resource>\n");
    }
  }
}

int mpp_resource_var(GsfInput *inp, GsfInput *inp2)
{
  int magic, i, item_count, data_size;
  guint8 buf[4];

  magic = mpp_read_int(inp);
  if (magic != 0xFADFADBA) {
    g_print("<!-- BAD MAGIC -->\n");
    return -1;
  }
  mpp_read_int(inp); /* unknown */
  item_count = mpp_read_int(inp);
  mpp_read_int(inp); /* unknown */
  mpp_read_int(inp); /* unknown */
  data_size = mpp_read_int(inp);

  for (i = 0;i < item_count;i ++) {
    int unique_id, type, offset;
    gsf_input_read(inp, 3, buf);
    buf[3] = 0;
    unique_id = GSF_LE_GET_GINT32(buf);
    type = (unsigned char)mpp_read_byte(inp);
    offset = mpp_read_int(inp);

    gsf_input_seek(inp2, offset, G_SEEK_SET);
    mpp_resource_var_data(inp2, type, unique_id);
  }
}

int mpp_assignment_fixed(GsfInput *inp, int itemSize,
	      GsfInput *inp2, int dataSize)
{
  int avail = gsf_input_remaining(inp);
  int magic = mpp_read_int(inp);
  int i, item_count;
  if (magic != 0xFADFADBA) {
    g_print("<!-- BAD MAGIC -->\n");
    return -1;
  }
  mpp_read_int(inp); /* unknown */
  item_count = mpp_read_int(inp);
  mpp_read_int(inp); /* unknown */
  item_count = (avail - 16) / itemSize;
  for (i = 0;i < item_count;i ++) {
    const guint8 *data = gsf_input_read(inp, itemSize, NULL);
    int size = GSF_LE_GET_GINT32(data);
    int offset = GSF_LE_GET_GINT32(data+4);

    size = dataSize;
    gsf_input_seek(inp2, offset, G_SEEK_SET);
    data = gsf_input_read(inp2, size, NULL);
    if (data) {
      g_print("<assignment idx=\"%d\">\n", i);
      mpp_assignment_fixed_data(data);
      g_print("</assignment>\n");
    }
  }
}

int mpp_assignment_var(GsfInput *inp, GsfInput *inp2)
{
  int magic, i, item_count, data_size;
  guint8 buf[4];
  guint8 *complete_hours = NULL, *incomplete_hours = NULL;

  magic = mpp_read_int(inp);
  if (magic != 0xFADFADBA) {
    g_print("<!-- BAD MAGIC -->\n");
    return -1;
  }
  mpp_read_int(inp); /* unknown */
  item_count = mpp_read_int(inp);
  mpp_read_int(inp); /* unknown */
  mpp_read_int(inp); /* unknown */
  data_size = mpp_read_int(inp);

  for (i = 0;i < item_count;i ++) {
    int unique_id, type, offset, size;
    gsf_input_read(inp, 3, buf);
    buf[3] = 0;
    unique_id = GSF_LE_GET_GINT32(buf);
    type = (unsigned char)mpp_read_byte(inp);
    offset = mpp_read_int(inp);

    gsf_input_seek(inp2, offset, G_SEEK_SET);
    size = mpp_read_int(inp2);
    if (type == 7) {
      mpp_assignment_complete_hours(gsf_input_read(inp2, size, NULL), unique_id);
    }
    else if (type == 9) {
      mpp_assignment_incomplete_hours(gsf_input_read(inp2, size, NULL), unique_id);
    }
  }
}

int main(int argc, char **argv)
{
  GsfInput *src;
  GsfInfile *infile;
  GsfInput *inp_fixed_meta, *inp_fixed_data, *inp_var_meta, *inp_var_data;
  GsfInfile *inf_project, *inf_task, *inf_constraint, *inf_resource, *inf_assignment;
  int i,n;

  if (argc != 2) {
    fprintf(stderr, "Usage: %s filename.mpp\n", argv[0]);
    return 255;
  }

  gsf_init();
  src = gsf_input_stdio_new (argv[1], NULL);
  if (!src) {
    fprintf(stderr, "Unable to open %s\n", argv[1]);
    return 255;
  }

  infile = gsf_infile_msole_new (src, NULL);
  if (!infile) {
    fprintf(stderr, "Unable to open %s\n", argv[1]);
    return 255;
  }

  if (!mpp_follow(infile, "Props9") ||
      !mpp_follow(infile, "   19/TBkndTask") ||
      !mpp_follow(infile, "   19/TBkndCons") ||
      !mpp_follow(infile, "   19/TBkndRsc") ||
      !mpp_follow(infile, "   19/TBkndAssn")) {
    fprintf(stderr, "Corrupted file\n");
    return 255;
  }

  g_print("<mpp version=\"9\">\n");

  /* Common properties */
  mpp_props(mpp_follow(infile, "Props9"));
  g_print("<summary>\n");
  mpp_summary(mpp_follow(infile, "\005SummaryInformation"));
  g_print("</summary>\n");
  g_print("<document-summary>\n");
  mpp_summary(mpp_follow(infile, "\005DocumentSummaryInformation"));
  g_print("</document-summary>\n");

  g_print("<project>\n");
  inf_project = GSF_INFILE(mpp_follow(infile, "   19"));

  /* Project properties */
  mpp_props(mpp_follow(inf_project, "Props"));

  /* Task information */
  inf_task = GSF_INFILE(mpp_follow(inf_project, "TBkndTask"));
  inp_fixed_meta = mpp_follow(inf_task, "FixedMeta");
  inp_fixed_data = mpp_follow(inf_task, "FixedData");
  struct mpp_task_fixed_meta *fixed_meta = mpp_task_fixed_meta_readahead(inp_fixed_meta, inp_fixed_data, 47, 768);
  int n_fixed_meta;
  for (n_fixed_meta = 0;fixed_meta[n_fixed_meta].idx != -1; n_fixed_meta++);
  qsort(fixed_meta, n_fixed_meta, sizeof(struct mpp_task_fixed_meta), mpp_task_fixed_meta_sort);

  inp_var_meta = mpp_follow(inf_task, "VarMeta");
  inp_var_data = mpp_follow(inf_task, "Var2Data");
  struct mpp_task_var_meta *var_meta = mpp_task_var_meta_readahead(inp_var_meta);
  int n_var_meta;
  for (n_var_meta = 0;var_meta[n_var_meta].idx != -1; n_var_meta++);
  qsort(var_meta, n_var_meta, sizeof(struct mpp_task_var_meta), mpp_task_var_meta_sort);

  mpp_task(fixed_meta, inp_fixed_meta, 47, inp_fixed_data, 768, var_meta, inp_var_data);

  g_free(fixed_meta);
  g_free(var_meta);

  /* Task constraint */
  inf_constraint = GSF_INFILE(mpp_follow(inf_project, "TBkndCons"));
  mpp_constraint(mpp_follow(inf_constraint, "FixedMeta"),
		 mpp_follow(inf_constraint, "FixedData"), 10, 20);

  /* Resource */
  inf_resource = GSF_INFILE(mpp_follow(inf_project, "TBkndRsc"));
  inp_fixed_meta = mpp_follow(inf_resource, "FixedMeta");
  inp_fixed_data = mpp_follow(inf_resource, "FixedData");
  inp_var_meta = mpp_follow(inf_resource, "VarMeta");
  inp_var_data = mpp_follow(inf_resource, "Var2Data");
  g_print("<resources>\n");
  mpp_resource_fixed(inp_fixed_meta, 37, inp_fixed_data);
  mpp_resource_var(inp_var_meta, inp_var_data);
  g_print("</resources>\n");

  /* Assignment */
  inf_assignment = GSF_INFILE(mpp_follow(inf_project, "TBkndAssn"));
  inp_fixed_meta = mpp_follow(inf_assignment, "FixedMeta");
  inp_fixed_data = mpp_follow(inf_assignment, "FixedData");
  inp_var_meta = mpp_follow(inf_assignment, "VarMeta");
  inp_var_data = mpp_follow(inf_assignment, "Var2Data");
  g_print("<assignments>\n");
  mpp_assignment_fixed(inp_fixed_meta, 34, inp_fixed_data, 142);
  mpp_assignment_var(inp_var_meta, inp_var_data);
  g_print("</assignments>\n");

  g_print("</project>\n");
  g_print("</mpp>\n");
}
